from __future__ import annotations

import textwrap
from pathlib import Path

import pytest

from nzb import Nzb, NzbMetaEditor

NZB_DIR = Path(__file__).parent.resolve() / "__nzbs__"
encoding = "utf-8"


def dedent(s: str) -> str:
    return textwrap.dedent(s).strip()


@pytest.mark.parametrize(
    "nzb_file",
    (
        NZB_DIR / "spec_example.nzb",
        NZB_DIR / "spec_example.nzb.gz",
    ),
    ids=lambda x: x.name,
)
def test_meta_clear(nzb_file: Path) -> None:
    edited = NzbMetaEditor.from_file(nzb_file).clear().to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_nzb_with_no_head_clear() -> None:
    nzb = NZB_DIR / "nzb_with_no_head.nzb"
    edited = NzbMetaEditor.from_file(nzb).clear().to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_remove_append() -> None:
    nzb = NZB_DIR / "spec_example.nzb"
    edited = NzbMetaEditor.from_file(nzb).remove("password").append(passwords="new secret!").to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="category">TV</meta>
            <meta type="password">new secret!</meta>
            <meta type="tag">HD</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_append_when_file_has_no_meta() -> None:
    nzb = NZB_DIR / "no_meta.nzb"
    edited = NzbMetaEditor.from_file(nzb).append(passwords="password appended").to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="password">password appended</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_append_when_file_has_single_meta() -> None:
    nzb = NZB_DIR / "single_meta.nzb"
    edited = NzbMetaEditor.from_file(nzb).append(tags="tag appended").to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">title</meta>
            <meta type="tag">tag appended</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_set_when_file_has_no_meta() -> None:
    nzb = NZB_DIR / "no_meta.nzb"
    edited = NzbMetaEditor.from_file(nzb).set(title="Big Buck Bunny").to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Big Buck Bunny</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_set_when_file_has_single_meta() -> None:
    nzb = NZB_DIR / "single_meta.nzb"
    edited = NzbMetaEditor.from_file(nzb).set(title="Big Buck Bunny").to_str()
    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Big Buck Bunny</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_remove_empty(tmp_path: Path) -> None:
    nzb = dedent("""
    <?xml version="1.0" encoding="utf-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)
    rm = NzbMetaEditor(nzb).remove("category").to_file(tmp_path / "test_meta_remove_empty.nzb")
    parsed = Nzb.from_file(rm)
    assert parsed.meta.title is None
    assert parsed.meta.passwords == ()
    assert parsed.meta.tags == ()
    assert parsed.meta.category is None


def test_meta_remove_one(tmp_path: Path) -> None:
    nzb = NZB_DIR / "single_meta.nzb"
    rm = NzbMetaEditor.from_file(nzb).remove("title").to_file(tmp_path / "single_meta.nzb")
    parsed = Nzb.from_file(rm)
    assert parsed.meta.title is None
    assert parsed.meta.passwords == ()
    assert parsed.meta.tags == ()
    assert parsed.meta.category is None


def test_meta_remove_missing_key(tmp_path: Path) -> None:
    nzb = NZB_DIR / "single_meta.nzb"
    rm = NzbMetaEditor.from_file(nzb).remove("akldakldjakldjs").to_file(tmp_path / "single_meta.nzb")
    parsed = Nzb.from_file(rm)
    assert parsed.meta.title is not None
    assert parsed.meta.passwords == ()
    assert parsed.meta.tags == ()
    assert parsed.meta.category is None


def test_meta_set(tmp_path: Path) -> None:
    nzb = dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)
    edited = NzbMetaEditor(nzb).set(title="New title", tags=["test", "test2"], passwords="New secret").to_str()

    assert edited == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">New title</meta>
            <meta type="category">TV</meta>
            <meta type="password">New secret</meta>
            <meta type="tag">test</meta>
            <meta type="tag">test2</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)


def test_meta_set_empty() -> None:
    nzb = dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="category">TV</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)
    out = NzbMetaEditor(nzb).set().to_str()  # no-op
    assert out == nzb


def test_meta_save_overwrite(tmp_path: Path) -> None:
    nzb = dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)

    nzb_file = tmp_path / "test_meta_save_overwrite.nzb"
    nzb_file.write_text(nzb, encoding="utf-8")

    edited = NzbMetaEditor(nzb).to_file(nzb_file, overwrite=True)
    assert nzb == edited.read_text("utf-8")


def test_no_doctype() -> None:
    # Since xml.etree.ElementTree does not preserve doctype,
    # we do our own hack on top to preserve it *if* it's
    # there, so we have this test to make sure we handle
    # the case where there is no doctype.
    original = dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """)
    edited = NzbMetaEditor(original).to_str()
    assert original == edited


def test_meta_editor_append_title() -> None:
    s = dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Big Buck Bunny - S01E01.mkv</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="John &lt;nzb@nowhere.example&gt;" date="1706440708" subject="[1/1] - &quot;Big Buck Bunny - S01E01.mkv&quot; yEnc (1/2) 1478616">
            <groups>
                <group>alt.binaries.boneless</group>
            </groups>
            <segments>
                <segment bytes="739067" number="1">9cacde4c986547369becbf97003fb2c5-9483514693959@example</segment>
                <segment bytes="739549" number="2">70a3a038ce324e618e2751e063d6a036-7285710986748@example</segment>
            </segments>
        </file>
    </nzb>
    """)
    editor = NzbMetaEditor(s).set(title="Big Buck Bunny", category="Movie").append(tags="1080p", passwords="secret2")
    assert editor.to_str() == dedent("""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Big Buck Bunny</meta>
            <meta type="category">Movie</meta>
            <meta type="password">secret</meta>
            <meta type="password">secret2</meta>
            <meta type="tag">HD</meta>
            <meta type="tag">1080p</meta>
        </head>
        <file poster="John &lt;nzb@nowhere.example&gt;" date="1706440708" subject="[1/1] - &quot;Big Buck Bunny - S01E01.mkv&quot; yEnc (1/2) 1478616">
            <groups>
                <group>alt.binaries.boneless</group>
            </groups>
            <segments>
                <segment bytes="739067" number="1">9cacde4c986547369becbf97003fb2c5-9483514693959@example</segment>
                <segment bytes="739549" number="2">70a3a038ce324e618e2751e063d6a036-7285710986748@example</segment>
            </segments>
        </file>
    </nzb>
    """)



"""
https://sabnzbd.org/wiki/extra/nzb-spec
https://web.archive.org/web/20240709113825/https://sabnzbd.org/wiki/extra/nzb-spec
"""  # noqa: D400, D415

from __future__ import annotations

import re
from datetime import datetime, timezone
from xml.etree import ElementTree

from nzb._exceptions import InvalidNzbError
from nzb._models import File, Meta, Segment


def parse_metadata(nzb: ElementTree.Element) -> Meta:
    """
    Parse the `<meta>...</meta>` field present in an NZB.

    ```xml
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
    </nzb>
    ```
    """
    passwords: list[str] = []
    tags: list[str] = []
    title = None
    category = None

    for meta in nzb.findall("./head/meta"):
        match meta.get("type", "").lower():
            case "title":
                title = meta.text

            case "password":
                # spec allows for multiple passwords by repeating the same field
                # <meta type="password">secret1</meta>
                # <meta type="password">secret2</meta>
                # <meta type="password">secret3</meta>
                if password := meta.text:
                    passwords.append(password)

            case "tag":
                # spec allows for multiple tags by repeating the same field
                # <meta type="tag">HD</meta>
                # <meta type="tag">Anime</meta>
                # <meta type="tag">1080p</meta>
                if tag := meta.text:
                    tags.append(tag.strip())

            case "category":
                category = meta.text
            case _:
                # Ignore unknown fields.
                pass

    return Meta(
        title=title,
        passwords=tuple(passwords),
        tags=tuple(tags),
        category=category,
    )


def parse_files(nzb: ElementTree.Element) -> tuple[File, ...]:
    """
    Parse the `<file>...</file>` field present in an NZB.

    ```xml
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>[...]</groups>
            <segments>[...]</segments>
        </file>
    </nzb>
    ```
    """  # noqa: E501
    files: list[File] = []

    for file in nzb.findall("./file"):
        try:
            poster = file.attrib["poster"]
            date = file.attrib["date"]
            subject = file.attrib["subject"]
        except KeyError as key:
            msg = f"Invalid or missing required attribute {key} in a 'file' element."
            raise InvalidNzbError(msg) from None

        try:
            posted_at = datetime.fromtimestamp(int(date), tz=timezone.utc)
        except ValueError:
            msg = "Invalid or missing required attribute 'date' in a 'file' element."
            raise InvalidNzbError(msg) from None

        groups = [group.text for group in file.findall("./groups/group") if group.text]
        groups.sort()
        if not groups:
            msg = (
                "Invalid or missing 'groups' element within the 'file' element. "
                "Each 'file' element must contain at least one valid 'groups' element."
            )
            raise InvalidNzbError(msg)

        segments: list[Segment] = []
        for segment in file.findall("./segments/segment"):
            try:
                size = int(segment.attrib["bytes"])
                number = int(segment.attrib["number"])
            except (KeyError, ValueError):
                # This segment is broken.
                # We do not error here because a few missing
                # segments don't invalidate the nzb.
                continue

            if message_id := segment.text:
                segments.append(Segment(size=size, number=number, message_id=message_id))
        segments.sort(key=lambda segment: segment.number)
        if not segments:
            msg = (
                "Invalid or missing 'segments' element within the 'file' element. "
                "Each 'file' element must contain at least one valid 'segments' element."
            )
            raise InvalidNzbError(msg)

        files.append(
            File(
                poster=poster,
                posted_at=posted_at,
                subject=subject,
                groups=tuple(groups),
                segments=tuple(segments),
            )
        )

    if not files:
        msg = (
            "Invalid or missing 'file' element in the NZB document. "
            "The NZB document must contain at least one valid 'file' element, "
            "and each 'file' must have at least one valid 'groups' and 'segments' element."
        )
        raise InvalidNzbError(msg)

    return tuple(sorted(files, key=lambda file: file.subject))


def generate_header(nzb: str) -> str:
    """
    Generate the two line header commonly seen in NZBs.

    ```xml
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    ```

    """
    header = '<?xml version="1.0" encoding="UTF-8"?>\n'
    if doctype := re.search(r"<!DOCTYPE nzb.*>", nzb, re.IGNORECASE):
        header += doctype.group().strip() + "\n"
    return header


def remove_header(xml: str) -> str:
    """
    Remove unnecessary headers.
    """
    xml = re.sub(r"^<\?xml\s+version.*?\?>", "", xml.strip(), flags=re.IGNORECASE)
    xml = re.sub(r"^<!DOCTYPE.*?>", "", xml.strip(), flags=re.IGNORECASE)
    xml = re.sub(r"^<nzb(\s*xmlns.*?)>", "<nzb>", xml.strip(), flags=re.IGNORECASE)
    return xml.strip()


def nzb_to_tree(nzb: str) -> ElementTree.Element:
    """
    Thin wrapper around ElementTree.fromstring for error handling.
    """
    try:
        return ElementTree.fromstring(remove_header(nzb))
    except ElementTree.ParseError:
        msg = "The NZB document is not valid XML and could not be parsed."
        raise InvalidNzbError(msg) from None



from __future__ import annotations

import json
import re
import textwrap
from pathlib import Path

import pytest

from nzb import InvalidNzbError, Nzb, NzbMetaEditor

NZB_DIR = Path(__file__).parent.resolve() / "__nzbs__"

INVALID_NZB_ERROR_GROUPS_ELEMENT = "Invalid or missing 'groups' element within the 'file' element. Each 'file' element must contain at least one valid 'groups' element."
INVALID_NZB_ERROR_SEGMENTS_ELEMENT = "Invalid or missing 'segments' element within the 'file' element. Each 'file' element must contain at least one valid 'segments' element."
INVALID_NZB_ERROR_FILE_ELEMENT = "Invalid or missing 'file' element in the NZB document. The NZB document must contain at least one valid 'file' element, and each 'file' must have at least one valid 'groups' and 'segments' element."

invalid_xml = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
        <meta type="password">secret</meta>
        <meta type="tag">HD</meta>
        <meta type="category">TV</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
        <segments>
            <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
            <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
        </segments>
    </file>
"""

valid_xml_but_invalid_nzb = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
    </file>
</nzb>"""


def test_invalid_nzb_error() -> None:
    try:
        message = "Missing something in the NZB"
        raise InvalidNzbError(message)
    except InvalidNzbError as error:
        assert error.message == message
        assert str(error) == message
        assert repr(error) == 'InvalidNzbError("Missing something in the NZB")'


def test_saving_overwrite() -> None:
    with pytest.raises(FileExistsError):
        file = NZB_DIR / "spec_example.nzb"
        NzbMetaEditor.from_file(file).to_file(file)


@pytest.mark.parametrize(
    "input_xml, expected_error",
    [
        pytest.param(
            invalid_xml,
            "The NZB document is not valid XML and could not be parsed.",
            id="truncated_xml",
        ),
        pytest.param(valid_xml_but_invalid_nzb, INVALID_NZB_ERROR_SEGMENTS_ELEMENT, id="missing_segments"),
    ],
)
def test_parsing_invalid_nzb(input_xml: str, expected_error: str) -> None:
    with pytest.raises(InvalidNzbError, match=expected_error):
        Nzb.from_str(input_xml)


def test_editing_invalid_nzb() -> None:
    with pytest.raises(
        InvalidNzbError,
        match="The NZB document is not valid XML and could not be parsed.",
    ):
        NzbMetaEditor(invalid_xml)


@pytest.mark.parametrize(
    "file_name, expected_error",
    (
        ("malformed_files.nzb", INVALID_NZB_ERROR_FILE_ELEMENT),
        ("malformed_files2.nzb", INVALID_NZB_ERROR_GROUPS_ELEMENT),
        ("malformed_groups.nzb", INVALID_NZB_ERROR_GROUPS_ELEMENT),
        ("malformed_segments.nzb", INVALID_NZB_ERROR_SEGMENTS_ELEMENT),
    ),
    ids=["malformed_files", "malformed_files2", "malformed_groups", "malformed_segments"],
)
def test_parser_exceptions(file_name: str, expected_error: str) -> None:
    with pytest.raises(InvalidNzbError, match=expected_error):
        Nzb.from_file(NZB_DIR / file_name)


def test_nzb_with_bad_file_date() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="not a date" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'date' in a 'file' element."):
        Nzb.from_str(nzb)


def test_nzb_with_missing_file_poster() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'poster' in a 'file' element."):
        Nzb.from_str(nzb)


def test_nzb_with_missing_file_subject() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'subject' in a 'file' element."):
        Nzb.from_str(nzb)


def test_non_existent_file() -> None:
    with pytest.raises(FileNotFoundError, match="blahblah"):
        Nzb.from_file("blahblah")


def test_nzb_from_json() -> None:
    data = json.dumps(
        {
            "meta": {"title": "Big Buck Bunny - S01E01.mkv", "passwords": ["secret"], "tags": ["HD"], "category": "TV"},
            "files": [
                {
                    "poster": "John <nzb@nowhere.example>",
                    "posted_at": "2024-01-28T11:18:28Z",
                    "subject": '[1/1] - "Big Buck Bunny - S01E01.mkv" yEnc (1/2) 1478616',
                    "groups": ["alt.binaries.boneless"],
                    "segments": [
                        {
                            "size": 739067,
                            "message_id": "9cacde4c986547369becbf97003fb2c5-9483514693959@example",
                        },
                        {
                            "size": 739549,
                            "number": 2,
                            "message_id": "70a3a038ce324e618e2751e063d6a036-7285710986748@example",
                        },
                    ],
                }
            ],
        }
    )
    with pytest.raises(
        InvalidNzbError,
        match=re.escape(
            "The provided JSON data is not in the expected format or contains invalid values. "
            "This method only works with JSON produced by `Nzb.to_json()`."
        ),
    ):
        Nzb.from_json(data)



from __future__ import annotations

import gzip
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath


def realpath(path: StrPath, /) -> Path:
    """
    Canonicalize a given path.
    """
    return Path(path).expanduser().resolve()


def read_nzb_file(path: StrPath, /) -> str:
    """
    Read a text file and return its content as a string.
    Handles both plain text and gzipped (.gz) files.
    """
    file = realpath(path)
    encoding = "utf-8"
    errors = "strict"

    if not file.is_file():
        raise FileNotFoundError(file)

    if file.suffix.casefold() == ".gz":
        # gzipped nzbs are fairly common (e.g., all of AnimeTosho)
        with gzip.open(file) as f:
            data = f.read().decode(encoding=encoding, errors=errors).strip()

    else:
        data = file.read_text(encoding=encoding, errors=errors)

    return data


def to_iterable(obj: str | Iterable[str] | None) -> Iterable[str]:
    if obj is None:
        return ()
    if isinstance(obj, str):
        return (obj,)
    if isinstance(obj, Iterable):
        return obj
    msg = f"Expected a string, an iterable of strings, or None, got {type(obj).__name__}"
    raise TypeError(msg)



from __future__ import annotations

import pytest

from nzb._utils import to_iterable


def test_to_iterable() -> None:
    assert to_iterable("hello") == ("hello",)
    assert to_iterable(None) == ()
    assert to_iterable(["a", "b", "c"]) == ["a", "b", "c"]

    with pytest.raises(TypeError):
        to_iterable(object())  # type: ignore[arg-type]



from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal, overload
from xml.etree import ElementTree

from nzb._exceptions import InvalidNzbError
from nzb._models import File, Meta, Segment
from nzb._parsers import generate_header, nzb_to_tree, parse_files, parse_metadata
from nzb._utils import read_nzb_file, realpath, to_iterable

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from pathlib import Path

    from _typeshed import StrPath, SupportsRichComparison
    from typing_extensions import Self


@dataclass(frozen=True, kw_only=True)
class Nzb:
    """
    Represents a complete NZB file.

    Example:
    -------
    ```python
    from nzb import Nzb

    text = '''
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Big Buck Bunny - S01E01.mkv</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="John &lt;nzb@nowhere.example&gt;" date="1706440708" subject="[1/1] - &quot;Big Buck Bunny - S01E01.mkv&quot; yEnc (1/2) 1478616">
            <groups>
                <group>alt.binaries.boneless</group>
            </groups>
            <segments>
                <segment bytes="739067" number="1">9cacde4c986547369becbf97003fb2c5-9483514693959@example</segment>
                <segment bytes="739549" number="2">70a3a038ce324e618e2751e063d6a036-7285710986748@example</segment>
            </segments>
        </file>
    </nzb>
    '''

    nzb = Nzb.from_str(text)

    print(f"{nzb.file.name} ({nzb.meta.category}) was posted by {nzb.file.poster} on {nzb.file.posted_at}.")
    print(f"Number of files: {len(nzb.files)}")
    print(f"Total size in bytes: {nzb.size}")
    ```

    """  # noqa: E501

    meta: Meta = field(default_factory=Meta)
    """Optional creator-definable metadata for the contents of the NZB."""

    files: tuple[File, ...]
    """File objects representing the files included in the NZB."""

    @classmethod
    def from_str(cls, nzb: str, /) -> Self:
        """
        Parse the given string into an [`Nzb`][nzb.Nzb].

        Parameters
        ----------
        nzb : str
            NZB string.

        Returns
        -------
        Nzb
            Object representing the parsed NZB file.

        Raises
        ------
        InvalidNzbError
            Raised if the NZB is invalid.

        """
        tree = nzb_to_tree(nzb)
        meta = parse_metadata(tree)
        files = parse_files(tree)

        return cls(meta=meta, files=files)

    @classmethod
    def from_file(cls, nzb: StrPath, /) -> Self:
        """
        Parse the given file into an [`Nzb`][nzb.Nzb].
        Handles both regular and gzipped NZB files.

        Parameters
        ----------
        nzb : StrPath
            Path to the NZB file.

        Returns
        -------
        Self
            Object representing the parsed NZB file.

        Raises
        ------
        FileNotFoundError
            Raised if the specified file doesn't exist.

        InvalidNzbError
            Raised if the NZB is invalid.

        """
        return cls.from_str(read_nzb_file(nzb))

    @classmethod
    def from_json(cls, data: str | bytes, /) -> Self:
        """
        Create an instance of this class from JSON data.

        Parameters
        ----------
        data : str | bytes
            JSON data representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        dejson = json.loads(data)

        try:
            meta = Meta(
                title=dejson["meta"]["title"],
                passwords=tuple(dejson["meta"]["passwords"]),
                tags=tuple(dejson["meta"]["tags"]),
                category=dejson["meta"]["category"],
            )
            files: list[File] = []

            for file in dejson["files"]:
                segments = tuple(
                    Segment(size=segment["size"], number=segment["number"], message_id=segment["message_id"])
                    for segment in file["segments"]
                )
                groups = tuple(file["groups"])
                posted_at = datetime.strptime(file["posted_at"], "%Y-%m-%dT%H:%M:%S%z")

                files.append(
                    File(
                        poster=file["poster"],
                        posted_at=posted_at,
                        subject=file["subject"],
                        groups=groups,
                        segments=segments,
                    )
                )
        except (KeyError, TypeError, ValueError):
            msg = (
                f"The provided JSON data is not in the expected format or contains invalid values. "
                f"This method only works with JSON produced by `{cls.__name__}.to_json()`."
            )
            raise InvalidNzbError(msg) from None

        return cls(meta=meta, files=tuple(files))

    def to_json(self, *, pretty: bool = False) -> str:
        """
        Serialize the instance of this class into a JSON string.

        Parameters
        ----------
        pretty : bool, optional
            Whether to pretty format the JSON string.

        Returns
        -------
        str
            JSON string representing this class.

        """

        def default(obj: Any) -> Any:
            if isinstance(obj, datetime):
                return obj.strftime("%Y-%m-%dT%H:%M:%SZ")
            msg = f"Object of type {type(obj).__name__} is not JSON serializable"  # pragma: no cover
            raise TypeError(msg)  # pragma: no cover

        indent = 2 if pretty else None
        return json.dumps(dataclasses.asdict(self), indent=indent, default=default)

    @cached_property
    def file(self) -> File:
        """
        The main content file (episode, movie, etc) in the NZB.
        This is determined by finding the largest non par2 file in the NZB
        and may not always be accurate.
        """
        return max((file for file in self.files if not file.is_par2()), key=lambda file: file.size)

    @cached_property
    def size(self) -> int:
        """Total size of all the files in the NZB."""
        return sum(file.size for file in self.files)

    @cached_property
    def filenames(self) -> tuple[str, ...]:
        """
        Tuple of unique file names across all the files in the NZB.
        May return an empty tuple if it fails to extract the name for every file.
        """
        return tuple(sorted({file.name for file in self.files if file.name is not None}))

    @cached_property
    def posters(self) -> tuple[str, ...]:
        """
        Tuple of unique posters across all the files in the NZB.
        """
        return tuple(sorted({file.poster for file in self.files}))

    @cached_property
    def groups(self) -> tuple[str, ...]:
        """
        Tuple of unique groups across all the files in the NZB.
        """
        groupset: set[str] = set()

        for file in self.files:
            groupset.update(file.groups)

        return tuple(sorted(groupset))

    @cached_property
    def par2_files(self) -> tuple[File, ...]:
        """
        Tuple of par2 files in the NZB.
        """
        return tuple(sorted((file for file in self.files if file.is_par2()), key=lambda f: f.subject))

    @cached_property
    def par2_size(self) -> int:
        """
        Total size of all the `.par2` files.
        """
        return sum(file.size for file in self.par2_files)

    @cached_property
    def par2_percentage(self) -> float:
        """
        Percentage of the size of all the `.par2` files relative to the total size.
        """
        return (self.par2_size / self.size) * 100

    def has_extension(self, ext: str, /) -> bool:
        """
        Check if any file in the NZB has the specified extension.

        This method ensures consistent extension comparison
        by normalizing the extension (removing any leading dot)
        and handling case-folding.

        Parameters
        ----------
        ext : str
            Extension to check for, with or without a leading dot (e.g., `.txt` or `txt`).

        Returns
        -------
        bool
            `True` if any file in the NZB has the specified extension, `False` otherwise.
        ```

        """
        return any(f.has_extension(ext) for f in self.files)

    def has_rar(self) -> bool:
        """
        Return `True` if any file in the NZB is a `.rar` file, `False` otherwise.
        """
        return any(file.is_rar() for file in self.files)

    def is_rar(self) -> bool:
        """
        Return `True` if all files in the NZB are `.rar` files, `False` otherwise.
        """
        return all(file.is_rar() for file in self.files)

    def is_obfuscated(self) -> bool:
        """
        Return `True` if any file in the NZB is obfuscated, `False` otherwise.
        """
        return any(file.is_obfuscated() for file in self.files)

    def has_par2(self) -> bool:
        """
        Return `True` if there's at least one `.par2` file in the NZB, `False` otherwise.
        """
        return any(file.is_par2() for file in self.files)


class NzbMetaEditor:
    def __init__(self, nzb: str, /) -> None:
        """
        Create an NzbMetaEditor instance.

        Parameters
        ----------
        nzb : str
            NZB content as a string.

        Raises
        ------
        InvalidNzbError
            If the string cannot be parsed as valid XML.

        Note
        ----
        This does not validate the given NZB.

        Example
        -------
        ```python
        from nzb import NzbMetaEditor

        text = '''
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
        <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
            <head>
                <meta type="title">Big Buck Bunny - S01E01.mkv</meta>
                <meta type="password">secret</meta>
                <meta type="tag">HD</meta>
                <meta type="category">TV</meta>
            </head>
            <file poster="John &lt;nzb@nowhere.example&gt;" date="1706440708" subject="[1/1] - &quot;Big Buck Bunny - S01E01.mkv&quot; yEnc (1/2) 1478616">
                <groups>
                    <group>alt.binaries.boneless</group>
                </groups>
                <segments>
                    <segment bytes="739067" number="1">9cacde4c986547369becbf97003fb2c5-9483514693959@example</segment>
                    <segment bytes="739549" number="2">70a3a038ce324e618e2751e063d6a036-7285710986748@example</segment>
                </segments>
            </file>
        </nzb>
        '''

        editor = NzbMetaEditor(text)
        edited = editor.set(title="Big Buck Bunny").append(tags="1080p").to_str()
        print(edited)
        ```

        """  # noqa: E501
        self._nzb = nzb
        self._tree = nzb_to_tree(self._nzb)

    @classmethod
    def from_file(cls, nzb: StrPath, /) -> Self:
        """
        Create an NzbMetaEditor instance from an NZB file path.

        Parameters
        ----------
        nzb : StrPath
            File path to the NZB.

        Returns
        -------
        Self
            Returns itself.

        """
        return cls(read_nzb_file(nzb))

    def sort(self, key: Callable[[ElementTree.Element], SupportsRichComparison] | None = None) -> Self:
        """
        Sort the metadata fields.

        Parameters
        ----------
        key : Callable[[ElementTree.Element], SupportsRichComparison], optional
            A callable that takes a `<meta>` Element and returns a comparable
            value.

        Returns
        -------
        Self
            Returns itself.

        Examples
        --------
        Here's an example where we sort the meta fields by their type.
        This is also what `sort()` does when you call it without a key.
        ```py
        editor = NzbMetaEditor(...)

        def key(element: ElementTree.Element) -> int:
            if typ := element.get("type"):
                return {"title": 0, "category": 1, "password": 2, "tag": 3}.get(typ, -1)
            return -1

        editor.sort(key=key)
        ```

        """
        if key is None:

            def key(element: ElementTree.Element) -> int:  # pragma: no cover
                if typ := element.get("type"):
                    return {"title": 0, "category": 1, "password": 2, "tag": 3}.get(typ, -1)
                return -1

        if (head := self._tree.find("./head")) is not None:
            head[:] = sorted(
                head.findall("./meta"),
                key=key,
            )
        return self

    def set(
        self,
        *,
        title: str | None = None,
        passwords: Iterable[str] | str | None = None,
        tags: Iterable[str] | str | None = None,
        category: str | None = None,
    ) -> Self:
        """
        Set metadata fields in the NZB.
        Provided fields are replaced entirely if they already exist.
        Fields that aren't provided remain unchanged.

        Parameters
        ----------
        title : str, optional
            The title metadata field.
        passwords : Iterable[str] | str, optional
            Password(s) for the NZB file.
        tags : Iterable[str] | str, optional
            Tag(s) associated with the NZB file.
        category : str, optional
            Category of the NZB file.

        Returns
        -------
        Self
            Returns itself.

        """

        head = self._tree.find("./head")
        if head is None:
            head = ElementTree.Element("head")
            self._tree.insert(0, head)

        if title:
            self.remove("title")
            sub = ElementTree.SubElement(head, "meta")
            sub.set("type", "title")
            sub.text = title

        if passwords:
            self.remove("password")
            for password in to_iterable(passwords):
                sub = ElementTree.SubElement(head, "meta")
                sub.set("type", "password")
                sub.text = password

        if tags:
            self.remove("tag")
            for tag in to_iterable(tags):
                sub = ElementTree.SubElement(head, "meta")
                sub.set("type", "tag")
                sub.text = tag

        if category:
            self.remove("category")
            sub = ElementTree.SubElement(head, "meta")
            sub.set("type", "category")
            sub.text = category

        return self.sort()

    def append(
        self,
        *,
        passwords: Iterable[str] | str | None = None,
        tags: Iterable[str] | str | None = None,
    ) -> Self:
        """
        Append metadata fields to the existing metadata in the NZB.

        Parameters
        ----------
        passwords : Iterable[str] | str, optional
            Password(s) for the NZB file.
        tags : Iterable[str] | str, optional
            Tag(s) associated with the NZB file.

        Returns
        -------
        Self
            Returns itself.

        """

        head = self._tree.find("./head")
        if head is None:
            head = ElementTree.Element("head")
            self._tree.insert(0, head)

        if passwords:
            for password in to_iterable(passwords):
                sub = ElementTree.SubElement(head, "meta")
                sub.set("type", "password")
                sub.text = password

        if tags:
            for tag in to_iterable(tags):
                sub = ElementTree.SubElement(head, "meta")
                sub.set("type", "tag")
                sub.text = tag

        return self.sort()

    @overload
    def remove(self, key: Literal["title", "password", "tag", "category"]) -> Self: ...

    @overload
    def remove(self, key: str) -> Self: ...

    def remove(self, key: Literal["title", "password", "tag", "category"] | str) -> Self:
        """
        Remove a metadata field from the NZB.
        If the same field is present multiple times, this will remove them all.

        Parameters
        ----------
        key : Literal["title", "password", "tag", "category"] | str, optional
            The metadata field to remove.

        Returns
        -------
        Self
            Returns itself.

        """

        if (head := self._tree.find("./head")) is not None:
            matches = [meta for meta in head.findall("./meta") if meta.get("type") == key]

            for match in matches:
                head.remove(match)

        return self

    def clear(self) -> Self:
        """
        Clear all metadata fields from the NZB.

        Returns
        -------
        Self
            Returns itself.

        """
        if (head := self._tree.find("./head")) is not None:
            head.clear()
            self._tree.remove(head)
        return self

    def to_str(self) -> str:
        """
        Return the edited NZB as a string.

        Returns
        -------
        str
            Edited NZB.

        """
        ElementTree.indent(self._tree, space=" " * 4)
        body: str = ElementTree.tostring(
            self._tree,
            encoding="utf-8",
            xml_declaration=False,
        ).decode("utf-8")
        body = body.replace("<nzb>", '<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">')
        return generate_header(self._nzb) + body

    def to_file(self, filename: StrPath, *, overwrite: bool = False) -> Path:
        """
        Save the edited NZB to a file.

        Parameters
        ----------
        filename : StrPath, optional
            Destination path for saving the NZB.
            This will also create the path if it doesn't exist already.
        overwrite : bool, optional
            Whether to overwrite the file if it exists, defaults to `False`.

        Returns
        -------
        Path
            The path to the saved file.

        Raises
        ------
        FileExistsError
            If the file exists and overwrite is `False`.

        """

        outfile = realpath(filename)

        if outfile.is_file() and not overwrite:
            raise FileExistsError(outfile)

        outfile.parent.mkdir(parents=True, exist_ok=True)
        outfile.write_text(self.to_str(), encoding="utf-8")
        return outfile



from __future__ import annotations

import pytest

from nzb._subparsers import extract_filename_from_subject, name_is_par2, name_is_rar


def test_name_is_rar() -> None:
    assert name_is_rar("") is False
    assert name_is_rar("blah") is False
    assert name_is_rar("blah.rar") is True
    assert name_is_rar("blah.par") is False


def test_name_is_par2() -> None:
    assert name_is_par2("") is False
    assert name_is_par2("blah") is False
    assert name_is_par2("blah.rar") is False
    assert name_is_par2("blah.par2") is True


@pytest.mark.parametrize(
    ("subject", "filename"),
    (
        (
            "[011/116] - [AC-FFF] Highschool DxD BorN - 02 [BD][1080p-Hi10p] FLAC][Dual-Audio][442E5446].mkv yEnc (1/2401) 1720916370",
            "[AC-FFF] Highschool DxD BorN - 02 [BD][1080p-Hi10p] FLAC][Dual-Audio][442E5446].mkv",
        ),
        (
            "[010/108] - [SubsPlease] Ijiranaide, Nagatoro-san - 02 (1080p) [6E8E8065].mkv yEnc (1/2014) 1443366873",
            "[SubsPlease] Ijiranaide, Nagatoro-san - 02 (1080p) [6E8E8065].mkv",
        ),
    ),
)
def test_extract_filename_from_subject(subject: str, filename: str) -> None:
    assert extract_filename_from_subject(subject) == filename



from __future__ import annotations


class InvalidNzbError(Exception):
    """Invalid NZB."""

    def __init__(self, message: str) -> None:
        self.message = message
        """Human readable error message."""
        super().__init__(message)

    def __str__(self) -> str:
        """Equivalent to accessing the .message attribute."""
        return self.message

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.message}")'



from __future__ import annotations

from dataclasses import dataclass
from os.path import splitext
from typing import TYPE_CHECKING

from nzb._subparsers import extract_filename_from_subject, name_is_par2, name_is_rar, stem_is_obfuscated

if TYPE_CHECKING:
    from datetime import datetime


@dataclass(frozen=True, kw_only=True, slots=True)
class Meta:
    """Optional creator-definable metadata for the contents of the NZB."""

    title: str | None = None
    """Title."""
    passwords: tuple[str, ...] = ()
    """Password(s)."""
    tags: tuple[str, ...] = ()
    """Tag(s)."""
    category: str | None = None
    """Category."""


@dataclass(frozen=True, kw_only=True, slots=True)
class Segment:
    """One part segment of a file."""

    size: int
    """Size of the segment."""
    number: int
    """Number of the segment."""
    message_id: str
    """Message ID of the segment."""


@dataclass(frozen=True, kw_only=True, slots=True)
class File:
    """Represents a complete file, consisting of segments that make up a file."""

    poster: str
    """The poster of the file."""
    posted_at: datetime
    """The date and time when the file was posted, in UTC."""
    subject: str
    """The subject of the file."""
    groups: tuple[str, ...]
    """Groups that reference the file."""
    segments: tuple[Segment, ...]
    """Segments that make up the file."""

    @property
    def size(self) -> int:
        """Size of the file calculated from the sum of segment sizes."""
        return sum(segment.size for segment in self.segments)

    @property
    def name(self) -> str | None:
        """
        Complete name of the file with it's extension extracted from the subject.
        May return `None` if it fails to extract the name.
        """
        return extract_filename_from_subject(self.subject)

    @property
    def stem(self) -> str | None:
        """
        Base name of the file without it's extension extracted from the [`File.name`][nzb._models.File.name].
        May return `None` if it fails to extract the stem.
        """
        if not self.name:
            return None
        root, _ = splitext(self.name)
        return root

    @property
    def extension(self) -> str | None:
        """
        Extension of the file without the leading dot extracted from the [`File.name`][nzb._models.File.name].
        May return `None` if it fails to extract the extension.
        """
        if not self.name:
            return None
        _, ext = splitext(self.name)
        return ext.removeprefix(".")

    def has_extension(self, ext: str, /) -> bool:
        """
        Check if the file has the specified extension.

        This method ensures consistent extension comparison
        by normalizing the extension (removing any leading dot)
        and handling case-folding.

        Parameters
        ----------
        ext : str
            Extension to check for, with or without a leading dot (e.g., `.txt` or `txt`).

        Returns
        -------
        bool
            `True` if the file has the specified extension, `False` otherwise.

        Examples
        --------
        ```python
        >>> file.has_extension('.TXT')  # True for 'file.txt'
        True
        >>> file.has_extension('txt')   # Also True for 'file.txt'
        True
        ```

        """
        if not self.extension:
            return False
        return self.extension.casefold() == ext.casefold().removeprefix(".")

    def is_par2(self) -> bool:
        """
        Return `True` if the file is a `.par2` file, `False` otherwise.
        """
        return False if self.name is None else name_is_par2(self.name)

    def is_rar(self) -> bool:
        """
        Return `True` if the file is a `.rar` file, `False` otherwise.
        """
        return False if self.name is None else name_is_rar(self.name)

    def is_obfuscated(self) -> bool:
        """
        Return `True` if the file is obfuscated, `False` otherwise.
        """
        # If we can't even determine the stem, it's most certainly obfuscated.
        return True if self.stem is None else stem_is_obfuscated(self.stem)



from __future__ import annotations

from nzb._core import Nzb, NzbMetaEditor
from nzb._exceptions import InvalidNzbError
from nzb._models import File, Meta, Segment

__all__ = (
    "File",
    "InvalidNzbError",
    "Meta",
    "Nzb",
    "NzbMetaEditor",
    "Segment",
)
